@@ -962,8 +962,8 @@ def test_params():
 
     # these are the initial param states
     expected = (
-        Param(0, "x", 1.0, 3.0, None, False, True, False, False, False, None, None),
-        Param(1, "y", 2.0, 4.0, None, False, False, True, False, True, None, 10),
+        Param(0, "x", 1.0, 3.0, None, False, True, None, None),
+        Param(1, "y", 2.0, 4.0, None, False, False, None, 10),
     )
     assert m.params == expected
 
@@ -1081,12 +1081,13 @@ def test_hesse_without_migrad():
     m.values["x"] = 1
     m.hesse()
     assert m.errors["x"] == approx((1.0 / 14.0) ** 0.5, abs=1e-4)
-    assert m.fmin is None
+    assert m.fmin
 
     m = Minuit(lambda x: 0, 0)
     m.errordef = 1
-    with pytest.raises(RuntimeError):
-        m.hesse()
+    m.hesse()
+    assert not m.accurate
+    assert m.fmin.hesse_failed
 
 
 def test_edm_goal():
@@ -1321,10 +1322,16 @@ def test_tolerance():
     m.migrad()
     assert m.valid
     edm = m.fmin.edm
+    m.tol = 0
     m.reset()
-    m.tol = 1e-6
     m.migrad()
     assert m.fmin.edm < edm
+    m.reset()
+    m.tol = None
+    assert m.tol == 0.1
+    m.reset()
+    m.migrad()
+    assert m.fmin.edm == edm
 
 
 def test_bad_tolerance():
@@ -1333,9 +1340,6 @@ def test_bad_tolerance():
     with pytest.raises(ValueError):
         m.tol = -1
 
-    with pytest.raises(ValueError):
-        m.tol = 0
-
 
 def test_issue_544():
     m = Minuit(func0, x=0, y=0)
@@ -1489,3 +1493,48 @@ def test_missing_ndata():
 
     m = Minuit(fcn, 1)
     assert_equal(m.ndof, np.nan)
+
+
+def test_issue_648():
+    class F:
+        errordef = 1
+        first = True
+
+        def __call__(self, a, b):
+            if self.first:
+                assert a == 1.0 and b == 2.0
+                self.first = False
+            return a ** 2 + b ** 2
+
+    m = Minuit(F(), a=1, b=2)
+    m.fixed["a"] = False  # this used to change a to b
+    m.migrad()
+
+
+def test_call_limit_reached_in_hesse():
+    m = Minuit(lambda x: ((x - 1.2) ** 4).sum(), np.ones(10) * 10)
+    m.errordef = 1
+    m.migrad(ncall=200)
+    assert m.fmin.has_reached_call_limit
+    assert m.fmin.nfcn < 205
+
+
+def test_issue_643():
+    def fcn(x, y, z):
+        return (x - 2) ** 2 + (y - 3) ** 2 + (z - 4) ** 2
+
+    fcn.errordef = Minuit.LEAST_SQUARES
+
+    m = Minuit(fcn, x=2, y=3, z=4)
+    m.migrad()
+
+    m2 = Minuit(fcn, x=m.values["x"], y=m.values["y"], z=m.values["z"])
+    # used to call MnHesse when it was not needed and quickly exhaust call limit
+    for i in range(10):
+        m2.minos()
+
+    m2.reset()
+    # used to exhaust call limit, because calls to MnHesse did not reset call count
+    for i in range(10):
+        m2.values = m.values
+        m2.minos()
